{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Author:       Angus Robertson, Magenta Systems Ltd
Description:  The ICS Application Monitor server is designed to monitor
              ICS applications using the TIcsAppMonCli client component,
              and ensure they remain running, restarting the application
              if it stops or becomes non-responsive, or on demand.
              Primarily to keep ICS server Windows services running
              non-stop, but may also be used for network wide monitoring
              of ICS applications.
Creation:     Sept 2024
Updated:      Sept 2024
Version:      V9.3
EMail:        francois.piette@overbyte.be  http://www.overbyte.be
Support:      https://en.delphipraxis.net/forum/37-ics-internet-component-suite/
Legal issues: Copyright (C) 2024 by Angus Robertson, Magenta Systems Ltd,
              Croydon, England. delphi@magsys.co.uk, https://www.magsys.co.uk/delphi/

              This software is provided 'as-is', without any express or
              implied warranty.  In no event will the author be held liable
              for any  damages arising from the use of this software.

              Permission is granted to anyone to use this software for any
              purpose, including commercial applications, and to alter it
              and redistribute it freely, subject to the following
              restrictions:

              1. The origin of this software must not be misrepresented,
                 you must not claim that you wrote the original software.
                 If you use this software in a product, an acknowledgment
                 in the product documentation would be appreciated but is
                 not required.

              2. Altered source versions must be plainly marked as such, and
                 must not be misrepresented as being the original software.

              3. This notice may not be removed or altered from any source
                 distribution.

              4. You must register this software by sending a picture postcard
                 to the author. Use a nice stamp and mention your name, street
                 address, EMail address and any comment you like to say.


DD Service Application Framework Requirement
--------------------------------------------
DDService is an enhanced Windows NT service application framework for Delphi and C++
Builder based on the original VCL service framework. In addition to it also encapsulates
new Windows NT service APIs introduced since Windows 2000. Original author was the late
Arno Garrels of Duodata in Berlin. Now being maintained by Magenta Systems Ltd.

https://www.magsys.co.uk/delphi/ddservice.asp

DDService must be installed before opening this sample.  Once the sample is built, it
may be installed as a Windows service from an elevated command line prompt with the
command IcsAppMon.exe /install. But this sample is written so it may also be run as
a GUI for debugging.


Windows Service or GUI?
-----------------------
If this sample is only used for monitoring, it does not matter how it is used.
But for Non-Stop monitoring of applications that requires the stopping and
restarting of applications, administative level access is required to start
and stop Windows services, and Windows services can not easily start GUI
applications.  So effectively, this program needs to be installed as a
Windows service to monitor other Windows Services, which is how most ICS
server applications should be designed, for continual running.


Application Independent Monitoring
----------------------------------
The ICS Application Monitor server is designed to monitor any ICS applications
using the TIcsAppMonCli client component, no configuration of the server is
needed other than setting it's listening IP address and email account details,
and the server has no prior knowledge of applications that may connect to it.

The server broadcasts it's availability, IP address and port to clients by
three methods:  Windows HLM registry entries, a named Windows message, and
optionally a UDP broadcast message.  The Windows registry and named messages
are only valid for applications on the same PC as the server, UDP messages
can be broadcast to the local LAN.

When clients access the server by the broadcasted IP address, the first send
a HELLO request, containing the required operating mode, currently 'Monitor
Only', 'Non-Stop Monitor' or 'Installation', and provide all the information
the server needs, such as application name, executable file name, Windows
Service name, Windows Handle, Process ID, etc, and general information the
client can add for information, and the client then keeps the server socket
connection open.  Note Non-Stop and Installation modes are only accepted
it the application is running on the same PC as the server, and the
server has the system rights to control the application.

Once accepted for monitoring, the clients sends a small PING packet once a
minute so the server knows it's running OK, with a STOP packet sent if
monitoring should be stopped.  If the client wants to be deliberately
restarted, for instance after an unexpected exception or terminal error,
it sends a special STOP packet that initiates an application restart
process.

If the client connection is closed without a STOP packet or the PING packets
stop for a while, the server will attempt to restart the client, by first
stopping the Windows service or program, waiting for it to stop, and then
starting the service or program again.  If the Windows service does not
stop cleanly due to being non-responsive, the server will attempt to
terminate the program by process ID to ensure it stops.

When sending packets, the client may add an email message that the server
will send to the admin email address, perhaps application start-up or close
down information, error information, anything useful really.  The server
also sends admin emails when clients start and stop monitoring.

Not implemented yet, but installation mode is similar to restarting, except
the exe file is replaced by a new version while being restarted, allowing
the application to update itself with a new version.  Longer term, there
may be application updating component to include downlaoding new versions
and support to install multiple files.

The server can potentially monitor an unlimited number of applications,
whose status will be available (soon) on a continuously refreshed web
page using websockets.   There is a USERINFO command allowing the client
and server to exchange application defined information.


Sep 2024 V9.3 - Baseline


This server and client are still under development and testing.  The final
version will extract the actual server parts of this application into a new
TIcsAppMonSrv server component that can be added to other applications for
more complex requirements.  But this sample is fully functional and being
tested on live public servers.

Pending - server becomes a component.
Pending - Web and WebSocket server to show status of monitored applications.
Pending - support for INSTALL command to install applications.
Pending - server and/or client send USERINFO to each other.



 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

unit IcsAppMonMain;

{$IF CompilerVersion < 15}
  {$MESSAGE FATAL 'This demo requires at least Delphi 7 or better'};
{$IFEND}

{$B-}                                 { Enable partial boolean evaluation   }
{$T-}                                 { Untyped pointers                    }
{$X+}                                 { Enable extended syntax              }
{$H+}                                 { Use long strings                    }
{$J+}                                 { Allow typed constant to be modified }

{$WARN UNSAFE_TYPE off}
{$WARN UNSAFE_CAST off}
{$WARN UNSAFE_CODE off}
{$WARN SYMBOL_PLATFORM OFF}
{$WARN SYMBOL_LIBRARY OFF}
{$WARN SYMBOL_DEPRECATED OFF}

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls, WinSvc,
  MagCommon,
  MagTaskWins,
  MagService,
//  MagHardwareEvents,
//  MagDBT,
  MagFirewall,
  OverbyteIcsIniFiles,
  OverbyteIcsUtils,
  OverbyteIcsTicks64,
  OverbyteIcsIpStreamLog,
  OverbyteIcsSslHttpOAuth,
  OverbyteIcsMailQueue,
  OverbyteIcsTypes,
  OverbyteIcsSslUtils,
  OverbyteIcsSslBase,
  OverbyteIcsWndControl,
  OverbyteIcsHttpSrv,
  OverbyteIcsHttpAppServer,
  OverbyteIcsAppMonCli;  // only for types and constants common to client and server

const
    ProgTitle = IcsAppMonName;
    ProgVersion = ProgTitle + ' - Release 1.0 - 18th September 2024' ;
    AppTitle = IcsAppMonName;
    SimpLogName = '"' + IcsAppMonName + '"yyyymmdd".log"' ;
    NameLogsPath = 'logs\' ;
    NameIniConfig = IcsAppMonName + '.config' ;
    EncodeKeyFileSrv = 61234 ;
    WM_FM_STARTUP = WM_USER + 804 ;

type
  TMonitorMain = class(TForm)
    MonitorServer: TIcsIpStrmLog;
    MonitorBroadcast: TIcsIpStrmLog;
    Panel1: TPanel;
    LogWin: TMemo;
    MagMailQueue: TIcsMailQueue;
    IcsRestEmail: TIcsRestEmail;
    TimerMain: TTimer;
    doClose: TButton;
    DiagLogPackets: TCheckBox;
    SslHttpAppSrv: TSslHttpAppSrv;
    procedure MagMailQueueOATokenEvent(ServNr: Integer; var Token, TokAccount: string; var TokExpireDT: TDateTime);
    procedure MagMailQueueLogEvent(LogLevel: TMailLogLevel; const Info: string);
    procedure FormCreate(Sender: TObject);
    procedure IcsRestEmailEmailNewToken(Sender: TObject);
    procedure IcsRestEmailEmailProg(Sender: TObject; LogOption: TLogOption; const Msg: string);
    procedure TimerMainTimer(Sender: TObject);
    procedure MonitorBroadcastLogProgEvent(Sender: TObject; Socnr: Integer; LogOption: TLogOption; const Msg: string);
    procedure MonitorBroadcastLogRecvEvent(Sender: TObject; Socnr: Integer; const Line: string);
    procedure MonitorServerLogProgEvent(Sender: TObject; Socnr: Integer; LogOption: TLogOption; const Msg: string);
    procedure MonitorServerLogRecvEvent(Sender: TObject; Socnr: Integer; const Line: string);
    procedure doCloseClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure MonitorServerLogChangeEvent(Sender: TObject; Socnr: Integer; LogState: TStrmLogState);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure HardwareEventsPowerEvent(Sender: TObject; PowerEvent: Integer; const Desc: string);
  private
    { Private declarations }
        procedure TriggerMessageEvent (var Msg: TMsg; var Handled: Boolean); virtual;
  public
    { Public declarations }
    procedure ApplicationEventsException(Sender: TObject; E: Exception);
    procedure AddDiagLine(S1: string);
    procedure AddLogLine (S1: string);
    procedure CloseLogFile;
    procedure FlushLogFile(OldFName: Boolean = False);
    procedure OpenLogFiles;
    procedure GetGenSettings ;
    procedure SendAdminEmail (const EmailTo, Subject, Body: string; Fatal: Boolean = False) ;
    procedure StartQueueMail;
    procedure StopQueueMail;
    procedure MonStartup;
    procedure MonStop;
    procedure WMFMSTARTUP (var Msg : TMessage); message WM_FM_STARTUP;
    procedure SendBroadcast;
    function AddSuppApp: Integer;
    function FindSuppApp(ClientID: Integer): Integer;
    procedure RemSuppAppId(ClientID: Integer);
    procedure RemSuppAppNr(AppNr: Integer);
    procedure ClrSuppAppNr(AppNr: Integer);
    function IsAppRunning(AppNr: Integer): Boolean;
    procedure AppStopOne(AppNr: Integer);
    procedure AppStartOne(AppNr: Integer);

  end;

var
    MonitorMain: TMonitorMain;
    GStartedByScm: Boolean = False;
    ServiceStartedByScm: string ;
    DiagLogBuffer: TIcsBuffLogStream ;     // temporary buffer for log
    FileIniFile: TIcsIniFile;
    FileIniConfig: String;
    ClosingFlag: Boolean;
    curDT: TDateTime;
    LogOpenFlag: Boolean;
    OldRefrToken: String;

// variables that will be part of component
// INI stuff
    FileLoggingsDir: string ;
    AdminEmailTo: string ;
    AdminEmailStart: boolean ;
    MonSrvIP: String;
    MonSrvPort: String;
    MonBcastIP: String;
    MonBcastPort: String;
    MonWebPort: String;

    FMonSuppApp: array of TIcsAMApp;
//    FHardwareEvents: THardwareEvents;
    FTotSuppApps: Integer;
    FMaxSuppApps: Integer;
    FWinMess_Query: Cardinal;
    FWinMess_Info: Cardinal;
    FTriggerBcast: Int64;
    FTriggerMinute: Int64;
    FAppFields: TStringList;
    FServiceCDatabaseHandle: SC_HANDLE = 0 ; // service control database
    FLANWide: Boolean = False;
    FIsAdmin: Boolean = False;
    FBCastSec: Integer;        // how often to broadcast our IP/port
    FDeadSecs: Integer;        // how long server waits before assuming application is dead, client controlled
    FMinDeadSecs: Integer;     // minimum wait before restarting server
    FStopWaitSecs : Integer;   // how long to wait for application to stop
    FStoppingFlag: Boolean = False;
    FStartWaitSecs: Integer;   // how long to wait before restarting application

implementation

{$R *.dfm}

procedure TMonitorMain.AddDiagLine (S1: string);
var
    res: integer ;
begin
    curDT := Now ;
    try
        S1 := TimeToStr (curDT) + IcsSpace + S1 ;
        if NOT GStartedByScm then
           LogWin.Lines.Add(S1);

   // file logging
        res := 0;
        try
            if Assigned (DiagLogBuffer) then
                res := DiagLogBuffer.WriteLine (S1) ;
        except
            res := -1 ;
        end ;
        if (res = -1) and Assigned (DiagLogBuffer) then
            DiagLogBuffer.WriteLine (S1) ;  // one retry before exception
    except
        IcsSimpleLogging (SimpLogName, 'Error Writing Log - ' + IcsGetExceptMess (ExceptObject)) ;
   end ;
end ;

procedure TMonitorMain.AddLogLine (S1: string);
begin
    AddDiagLine(S1);
end;

// Feb 2019 can now flush logs with old file name at midnight
procedure TMonitorMain.FlushLogFile(OldFName: Boolean = False) ;
begin
    try
        if Assigned (DiagLogBuffer) then
        begin
        //    if NOT FileInUse (DiagLogBuffer.FullName) then
                DiagLogBuffer.FlushFile(OldFName)
          //  else
         //       AddLogLine ('Skipped Flushing Diag Log, In-Use') ;
        end ;
    except
        IcsSimpleLogging (SimpLogName, 'Error Flushing Logs - ' + IcsGetExceptMess (ExceptObject)) ;
    end ;
end ;

procedure TMonitorMain.CloseLogFile ;
begin
    try
        if Assigned (DiagLogBuffer) then
        begin
            DiagLogBuffer.FlushFile ;
            DiagLogBuffer.Free ;
            DiagLogBuffer := Nil ;
        end ;
    except
        IcsSimpleLogging (SimpLogName, 'Aborting - Error Closing Diag Log') ;
    end ;
    LogOpenFlag := false ;
end ;


procedure TMonitorMain.doCloseClick(Sender: TObject);
begin
    Close;
end;

procedure TMonitorMain.OpenLogFiles ;
var
    diagfname, S: string;
begin
    curDT := Now ;
    CloseLogFile ;

    diagfname := IcsDQUOTE + FileLoggingsDir + '\' + IcsAppMonName + '-"yyyymmdd".log"' ;
    S := '"' + ProgVersion + IcsCRLF +
         'Log File Created: "dd mmm yyyy hh:mm:ss"' + IcsCRLF +
         'Computer: ' + IcsGetCompName + '"' + IcsCRLF ;
    try
        DiagLogBuffer := TIcsBuffLogStream.Create (Self, diagfname, S) ;
        DiagLogBuffer.FlushFile ;
        LogOpenFlag := true ;
    except
        IcsSimpleLogging (SimpLogName, 'Aborting - Error Opening Log - ' + IcsGetExceptMess (ExceptObject)) ;
    end ;
end ;

procedure TMonitorMain.GetGenSettings ;
var
    section: string ;
begin
    if NOT FileExists (FileIniConfig) then exit ;
    FreeAndNil (FileIniFile) ;
    try
    FileIniFile := TIcsIniFile.Create (FileIniConfig) ;
    section := 'General' ;
    with FileIniFile do
    begin
        FileLoggingsDir := ReadString (section, 'FileLoggingsDir', 'd:\weblogs\' + IcsAppMonName + '\') ;
        AdminEmailTo := ReadString (section, 'AdminEmailTo', '') ;
        AdminEmailStart := IcsCheckTrueFalse (ReadString (section, 'AdminEmailStart', 'False')) ;
        MonSrvIP := ReadString (section, 'MonSrvIP', '0.0.0.0') ;
        MonSrvPort := ReadString (section, 'MonSrcPort', DefSrvPort) ;
        MonBcastIP := ReadString (section, 'MonBcastIP', '255.255.255.255') ;
        MonBcastPort := ReadString (section, 'MonBcastPort', DefBcastPort) ;
        MonWebPort := ReadString (section, 'MonWebPort', DefWebPort) ;
    end ;
    except
        IcsSimpleLogging (SimpLogName, 'Aborting - Failed to open INI File - ' + IcsGetExceptMess (ExceptObject)) ;
    end ;
end ;

procedure TMonitorMain.IcsRestEmailEmailNewToken(Sender: TObject);
var
    info: String;
begin
// we should save refresh token since only get it after a login window
    if (IcsRestEmail.RefrToken <> '') and (IcsRestEmail.RefrToken <> OldRefrToken) then begin
        info := 'New OAuth2 Refresh Token for Account: ' + IcsRestEmail.NewAccEmail +
                ', Should be Saved in Config File' + IcsCRLF +
                'RefrToken=' + IcsRestEmail.RefrToken;
        AddLogLine(info);
        if (AdminEmailTo <> '') then
            SendAdminEmail (AdminEmailTo, 'TelecomWebApp New OAuth2 Refresh Token',
                                                'Web server New OAuth2 Refresh Token.' + IcsCRLF + IcsCRLF + info);
    end;
end;

procedure TMonitorMain.IcsRestEmailEmailProg(Sender: TObject; LogOption: TLogOption; const Msg: string);
begin
    AddLogLine (Msg) ;
end;

procedure TMonitorMain.StartQueueMail;
var
    S: string ;
    I: integer ;
begin
    try
        if NOT MagMailQueue.Active then
        begin
            IcsLoadMailQuFromIni(FileIniFile, MagMailQueue, 'MailQueue');
            if (MagMailQueue.MailQuDir = '') or (NOT ForceDirectories (MagMailQueue.MailQuDir)) then
            begin
                AddLogLine ('!! Failed to Start Mail Queue, No Directory') ;
                exit ;
            end;
            IcsLoadRestEmailFromIni(FileIniFile, IcsRestEmail, 'RestEmail');  // June 2022
            OldRefrToken := IcsRestEmail.RefrToken;                           // June 2022
        end;
        if MagMailQueue.MailServers.Count = 0 then
        begin
            AddLogLine ('!! Failed to Start Mail Queue, No Mail Servers') ;
            exit ;
        end;
        S := '';
        for I := 0 to MagMailQueue.MailServers.Count - 1 do
        begin
            MagMailQueue.MailServers[I].Password := MagstrBXDecryptEx (MagMailQueue.MailServers[I].Password, EncodeKeyFileSrv) ;
            S := S + MagMailQueue.MailServers[I].Host + ', ';
        end;
        MagMailQueue.Debug := False;
        MagMailQueue.QuStartDelay := 0; // no delay sending email
        MagMailQueue.Active := true ;
        if MagMailQueue.Active then
        begin
             AddLogLine ('Started Mail Queue OK, Servers: ' + S) ;
        end;
    except
         AddLogLine ('!! Failed to Start Mail Queue - ' +  IcsGetExceptMess (ExceptObject)) ;
    end;
 end;

procedure TMonitorMain.StopQueueMail;
begin
    if MagMailQueue.Active then
    begin
        if (MagMailQueue.MailImmItems > 0) then
                 AddLogLine ('Waiting up to 30 seconds to send Queued Mail Items') ;
        MagMailQueue.WaitSendandStop (30) ;
    end ;
end;


procedure TMonitorMain.SendAdminEmail (const EmailTo, Subject, Body: string; Fatal: Boolean = False) ;   // Dec 2023 added Fatal
var
    sTempFrom: string ;
    id: integer ;
begin
    if Fatal then begin
        if NOT MagMailQueue.Active then exit;   // July 2024 only once
        AddDiagLine ('Stopping Mail Queue After Fatal Error') ;
        MagMailQueue.Active := False;
        exit;
    end
    else begin
        if NOT MagMailQueue.Active then
            StartQueueMail ;  // restart if not running
        if NOT MagMailQueue.Active then
        begin
             AddDiagLine ('Error Starting Mail Queue') ;
        end ;
    end;
    try
        with MagMailQueue.QuHtmlSmtp do
        begin

            sTempFrom := AppTitle + '_' + IcsGetCompName ;   // May 2024
            sTempFrom := '"' + sTempFrom + '" <' + sTempFrom + '@magsys.co.uk>' ;
            EmailFiles.Clear ;
            RcptName.clear;
            Allow8bitChars := true ;
            ContentType := smtpPlainText ;
            WrapMessageText := false ;
            WrapMsgMaxLineLen := 76 ;
            PlainText.Clear ;
            PlainText.Text := IcsAppMonTitle + ' on ' + IcsGetCompName + IcsCRLF + IcsCRLF + Body ;
            AddDiagLine ('Queuing Email Form to ' + EmailTo + ' from ' + sTempFrom) ;
            RcptName.Clear ;
            RcptName.Add (EmailTo) ;
            FromName := sTempFrom ;
            HdrCc := '' ;  // Aug 2014
            HdrTo := EmailTo ;
            HdrFrom := sTempFrom ;
            HdrReplyTo := sTempFrom ;
            HdrSubject := Subject ;
            XMailer := '' ;
            id := MagMailQueue.QueueMail ;
            if id > 0 then
            begin
       // done OK
                AddDiagLine ('Email Form Queued OK') ;
            end
            else
            begin
                AddDiagLine ('Failed to Queue Email Form: ' + ErrorMessage) ;
            end ;
        end ;
    except
        AddDiagLine ('Failed to Queue Email Form: ' + IcsGetExceptMess (ExceptObject)) ;
    end;
    FlushLogFile;  // May 2021
end ;


procedure TMonitorMain.MagMailQueueLogEvent(LogLevel: TMailLogLevel; const Info: string);
begin
    if LogLevel <= MLogLevelInfo then
        AddLogLine (Info) ;
end;

procedure TMonitorMain.MagMailQueueOATokenEvent(ServNr: Integer; var Token, TokAccount: string; var TokExpireDT: TDateTime);
begin
    AddLogLine('Starting to get OAuth2 Bearer Token');
    if NOT IcsRestEmail.GetNewToken(False) then   // no interaction, waits for browser window to be completed
        AddLogLine('Failed to get OAuth2 Bearer Token')
    else begin
        Token := IcsRestEmail.AccToken;
        TokExpireDT := IcsRestEmail.AccExpireDT;
        TokAccount := IcsRestEmail.NewAccEmail;
      // note, AccToken has a short life, few hours, no need to save it in INI file, it can be refreshed
    end;
end;



procedure TMonitorMain.ApplicationEventsException(Sender: TObject; E: Exception);
var
    S: string ;
begin
    S := '!!! Application Exception Event - ' + IcsGetExceptMess (E) ;
    if LogOpenFlag then
    begin
        AddLogLine (S) ;
        FlushLogFile;
    end
    else
        IcsSimpleLogging (SimpLogName, S) ;
end;

procedure TMonitorMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
    if ClosingFlag then exit ;
    ClosingFlag := true ;
    try
        TimerMain.Enabled := false ;
        if Assigned (DiagLogBuffer) then
            AddLogLine ('FormClose, Stopping Server Monitor') ;
        MonStop;
    except
    end ;
end;

procedure TMonitorMain.FormDestroy(Sender: TObject);
begin
    try
        if NOT ClosingFlag then
            MonStop;
        FreeAndNil (FileIniFile) ;
        CloseLogFile ;
//        FreeAndNil (FHardwareEvents);
        FreeAndNil (DiagLogBuffer) ;
        FAppFields.Free;
    except
    end;
end;

procedure TMonitorMain.FormCreate(Sender: TObject);
begin
    Caption := ProgVersion ;
    Application.OnMessage := TriggerMessageEvent ;
    Application.OnException := ApplicationEventsException ;
    if NOT MagCheckUniqueInstance then
    begin
        IcsSimpleLogging (SimpLogName, 'ICS Application Monitor is Already Running') ;
        Application.Terminate ;
        exit ;
    end ;
    FileIniConfig := ExtractFileDir (Lowercase (ParamStr(0))) + '\' + NameIniConfig ;   // main config file
    IcsSimpleLogging (SimpLogName, 'Config File: ' + FileIniConfig) ;
    GetGenSettings ;
    if FileLoggingsDir = '' then begin
        IcsSimpleLogging (SimpLogName, 'Fatal Error, No Config File') ;
        Application.Terminate;
        Exit;
    end;
    IcsSimpleLogging (SimpLogName, 'Logs Path: ' + FileLoggingsDir) ;
    FBCastSec := DefBCastSec;
    FDeadSecs := DefDeadSecs;
    FMinDeadSecs := DefMinDeadSecs;
    FStopWaitSecs := DefStopWaitSecs;
    FStartWaitSecs := DefStartWaitSecs;
    PostMessage (Handle, WM_FM_STARTUP, 0, 0) ;
end;

procedure TMonitorMain.WMFMSTARTUP (var Msg : TMessage);
begin
    MonStartup ;
end;

function TMonitorMain.AddSuppApp: Integer;
var
    I: Integer;
begin
    Result := -1;
    if FTotSuppApps >= FMaxSuppApps then begin
        FMaxSuppApps := FMaxSuppApps * 2;
        SetLength(FMonSuppApp, FMaxSuppApps);
    end;
    FTotSuppApps := FTotSuppApps + 1;
    for I := 0 to FMaxSuppApps - 1 do begin
        if (FMonSuppApp[I].AppClientID = 0) then begin
            Result := I;
            ClrSuppAppNr(I);
            exit;
        end;
    end;
end;

function TMonitorMain.FindSuppApp(ClientID: Integer): Integer;
var
    I: Integer;
begin
    Result := -1;
    if FTotSuppApps = 0 then
        Exit;
    for I := 0 to FMaxSuppApps - 1 do begin
        if (FMonSuppApp[I].AppClientID = ClientId) then begin
            Result := I;
            exit;
        end;
    end;
end;

procedure TMonitorMain.RemSuppAppId(ClientID: Integer);
var
    Nr: Integer;
begin
    Nr := FindSuppApp(ClientID);
    if Nr >= 0 then
        RemSuppAppNr(Nr);
end;

procedure TMonitorMain.RemSuppAppNr(AppNr: Integer);
begin
    if AppNr >= FMaxSuppApps then
        Exit;
    ClrSuppAppNr(AppNr);
    FTotSuppApps := FTotSuppApps - 1;
end;

procedure TMonitorMain.ClrSuppAppNr(AppNr: Integer);
begin
    with FMonSuppApp[AppNr] do begin
        AppClientTitle := 'New Remote Client';
        AppClientID := 0;
        AppExeFile := '';
        AppServiceName := 'NONE';
        AppWinHnd := 0;
        AppCompName := '';
        AppLocalIP := '';
        AppDeadSecs := 0;
        AppStarted := 0;
        AppLastStartTick := Trigger64Disabled;
        AppLastPingTick := Trigger64Disabled;
        AppLastPongTick := Trigger64Disabled;
        AppStoppingTick := Trigger64Disabled;
        AppSuppState := AppStateNone;
        AppRestartMode := RestModeAuto;
        AppStopState := StopStateNone;
        AppStartState := StartStateNone;
        AppEmailLines := '';
        AppEmailTick := Trigger64Disabled;
    end;
end;

procedure TMonitorMain.MonStartup;
var
    access: DWORD;
    S: String;
begin
    FStoppingFlag := False;
    try
        FMaxSuppApps := 32;
        FTotSuppApps := 0;
        SetLength(FMonSuppApp, FMaxSuppApps);
        FAppFields := TStringList.Create;
        FAppFields.Delimiter := ProtDelim;
        FAppFields.StrictDelimiter := True;

    // logging and OpenSSL for mail
        OpenLogFiles;
        AddLogLine ('');
        AddLogLine (ProgVersion);
        AddLogLine ('OS: ' + MagGetOSVersion);
        AddLogLine ('Compiler: ' + IcsBuiltWithEx) ;
        IcsSslRootCAStore.Initialise;
        AddLogLine('SSL/TLS ' + IcsReportOpenSSLVer(True));

    // Windows Service Manager
        access := SC_MANAGER_ALL_ACCESS ;
        if (NOT IsProgAdmin) then
            access := GENERIC_READ ;
        FServiceCDatabaseHandle := OpenSCManager (Nil , Nil, access) ;
        if FServiceCDatabaseHandle = 0 then
            AddLogLine ('Failed to Open Service Manager: ' + MagFormatLastError) ;

   // TCP/IP server for clients to talk to us
        MonitorServer.SrvIcsHosts.Clear;
        MonitorServer.SrvIcsHosts.Add;
        with MonitorServer.SrvIcsHosts[0] do begin
            BindIpAddr := MonSrvIP ;
            BindNonPort := atoi(MonSrvPort);
            Enabled := True;
            HostTag := 'MonServer' ;
            Descr := HostTag;
        end;
        MonitorServer.LogProtocol := logprotTcpServer;
        MonitorServer.MaxSockets := FMaxSuppApps + 5;
        MonitorServer.MaxLineLen := 4096;
        MonitorServer.AddCRLF := True;
        MonitorServer.StartLogging;
        AddLogLine ('Started Monitor Server on ' + MonSrvIP + ':' + MonSrvPort);

    // broadcast to clients to tell them who we are
        FWinMess_Query := RegisterWindowMessage (IcsAppMonMessQuery) ;
        FWinMess_Info := RegisterWindowMessage (IcsAppMonMessInfo) ;
        if FLANWide then begin
            MonitorBroadcast.RemoteHost := MonBcastIP;
            MonitorBroadcast.RemoteIpPort := MonBcastPort;
            MonitorBroadcast.LocalIpAddr := MonSrvIP;
            MonitorBroadcast.LocalIpPort := '0';
            MonitorBroadcast.LogProtocol := logprotUdpClient;
            MonitorBroadcast.AddCRLF := True;
            MonitorBroadcast.StartLogging;
            AddLogLine ('Started Monitor Broadcast to Port ' + MonBcastPort) ;
        end;
        FIsAdmin := IcsIsProgAdmin;
        if (MonSrvIP <> IcsRegGetHlm(IcsAppMonRegKey, IcsAppMonRegAddr)) or
                                    (MonSrvPort <> IcsRegGetHlm(IcsAppMonRegKey, IcsAppMonRegPort)) then begin
            if FIsAdmin then begin
                if NOT IcsRegPutHlm(IcsAppMonRegKey, IcsAppMonRegAddr, MonSrvIP) then
                    AddLogLine ('!! Faile to Update HLM Registry: ' + IcsAppMonRegKey) ;
                IcsRegPutHlm(IcsAppMonRegKey, IcsAppMonRegPort, MonSrvPort);
                IcsRegPutHlm(IcsAppMonRegKey, IcsAppMonRegDT, RFC3339_DateToStr(Now));  // ISO
            end
            else
                AddLogLine ('!! Program Needs Adminstrator Rights to Set Server IP Settings in Registry') ;
        end;

    // open firewall and add rule for our server - needs admin rights
       if FIsAdmin then begin
            try
                AddLogLine (Trim(MagFireWallRulesEnum('')));  // list settings
                S := MagFireWallRulesAdd(IcsAppMonTitle, 'Magsys', IcsAppMonTitle, ParamStr(0), FirewallBoth);
                AddLogLine ('Windows Firewall Rule: ' + S);
            except
                AddLogLine ('Failed to Access Firewall - ' + IcsGetExceptMess (ExceptObject)) ;
            end;
       end;

    // power events
//        FHardwareEvents := THardwareEvents.Create(Self);
//        FHardwareEvents.onPowerEvent := HardwareEventsPowerEvent;

    // mail queue
        FTriggerBcast := Trigger64Immediate;
        FTriggerMinute := Trigger64Immediate;
        StartQueueMail;
        if AdminEmailStart then
            SendAdminEmail (AdminEmailTo, IcsAppMonTitle, IcsAppMonTitle + ' Started OK, Listing on:' + MonSrvIP + ':' + MonSrvPort);
        TimerMain.Enabled := True;
    except
        AddDiagLine ('Failed to Start Server Monitor: ' + IcsGetExceptMess (ExceptObject)) ;
        IcsSimpleLogging (SimpLogName, 'Failed to Start Server Monitor: ' + IcsGetExceptMess (ExceptObject)) ;
        SendAdminEmail (AdminEmailTo, IcsAppMonTitle, IcsAppMonTitle + ' Failed to Start');
        Application.Terminate;
    end;
end;

procedure TMonitorMain.MonStop;
begin
    FStoppingFlag := True;
    TimerMain.Enabled := False;
    AddLogLine ('Stopping Client Monitoring and Servers');
    if AdminEmailStart then
        SendAdminEmail (AdminEmailTo, IcsAppMonTitle, IcsAppMonTitle + ' Stopping');
  // clear date in registry
    if FIsAdmin then
        IcsRegPutHlm(IcsAppMonRegKey, IcsAppMonRegDT, '');
    MonitorBroadcast.StopLogging;
    MonitorServer.StopLogging;
    if MagMailQueue.Active then
        StopQueueMail ;
    if FServiceCDatabaseHandle <> 0 then
        CloseServiceHandle (FServiceCDatabaseHandle) ;
end;

// tell other applications we are running, always local broadcast message to this PC, maybe remote UDP broadcast
procedure TMonitorMain.SendBroadcast;
var
    Packet: String;
    Mywp: WPARAM;
    Mylp: LPARAM;
    Success: Boolean;
begin
  // broadcast server IP to other applications on this desktop
    try
        Mywp := WPARAM(WSocketStrToIPv4(MonSrvIP, Success));  // returns integer!
        Mylp := atoi(MonSrvPort);
        if Mylp > 0 then begin
            PostMessage(HWND_BROADCAST, FWinMess_Info, Mywp, Mylp);
            if DiagLogPackets.Checked then
            AddLogLine ('Broadcast Server Info Message');
        end;
    except
        AddLogLine ('Post Message Exception');
    end;

 // UDP broadcast to other computers
    try
        if FLANWide and (MonitorBroadcast.LogActive) then begin
            Packet := IcsAppMonName + ProtDelim +     // BcastFieldServMon = 0
                      ProtCmdHello + ProtDelim +      // BcastFieldCommand = 1
                      IcsAppMonVerNr + ProtDelim +    // BcastFieldVersion = 2
                      IcsGetCompName + ProtDelim +    // BcastFieldSrvName = 3
                      MonSrvIP + ProtDelim +          // BcastFieldSrvIP = 4
                      MonSrvPort + ProtDelim;         // BcastFieldSrvPort = 5
            MonitorBroadcast.SendLogLine(Packet);
            if DiagLogPackets.Checked then
                AddLogLine ('UDP Broadcast: ' + Packet);
        end;
    except
        AddDiagLine ('Exception Broadcast: ' + IcsGetExceptMess (ExceptObject)) ;
    end;
end;

// is Windows service or program running (by Process ID)
// paused Windows service is considered running
function TMonitorMain.IsAppRunning(AppNr: Integer): Boolean;
var
    Resp: String;
begin
    Result := False;
    if AppNr >= FMaxSuppApps then
        Exit;
    try
        with FMonSuppApp[AppNr] do begin
            if (AppServiceName <> '') and (AppServiceName <> ServiceNone) then begin
                if FServiceCDatabaseHandle = 0 then
                    Exit;
                Result := MagCrtlServEx (FServiceCDatabaseHandle, AppServiceName, servfCheck, 0, Resp);  // no wait
             //   AddLogLine (AppServiceName + ' : ' + Resp);    // TEMP

             // check same service is still running, it might have restarted!
                if Result and (AppWinPID <> 0) then
                    Result := MagCheckPID (AppWinPID);
            end
            else begin
                if AppWinPID <> 0 then
                    Result := MagCheckPID (AppWinPID);
            end;
        end;
    except
        AddDiagLine ('Exception IsAppRunning: ' + IcsGetExceptMess (ExceptObject)) ;
    end;
end;

// stop or closing a Windows service or program, always exit after changing State
procedure TMonitorMain.AppStopOne(AppNr: Integer);
var
    Resp: String;
    StoppedSecs, exitcode: Integer;
    Running: Boolean;
//    ProcInfo: TProcessInformation;
begin
    if FStoppingFlag then
       Exit;
    if AppNr >= FMaxSuppApps then
        Exit;
    Running := IsAppRunning(AppNr);
    exitcode := 0;
    with FMonSuppApp[AppNr] do begin
    try
        StoppedSecs := IcsElapsedSecs64 (AppStoppingTick);
        if AppExeFile = '' then begin
            AddLogLine (AppClientTitle + ': Hello Failed, No EXE Filer');
            AppStartState := StartStateNone;
            Exit;
        end;

     // service needs to be stopped, program confirmed as not running, then service started again
     // beware, this program needs admin rights to start and stop Windows services, not under the debugger
        if FIsAdmin and (AppServiceName <> '') and (AppServiceName <> ServiceNone) then begin
            if FServiceCDatabaseHandle = 0 then begin
                AddLogLine (AppClientTitle + ': No Service Control Handle');
                AppStopState := StopStateNone;
                Exit;
            end;

        // stop windows service, if running
            if AppStopState = StopStateStopProg then begin
                AppStopState := StopStateStopWait;
                if Running then begin
                    AddLogLine(AppClientTitle + ': Stopping Windows Service - ' +  AppServiceName);
             // no wait or error checking, keep checking if service stops, don't want to block this app
                    MagCrtlServEx (FServiceCDatabaseHandle, AppServiceName, servfStop, 0, Resp);
                    AddLogLine (AppServiceName + ': ' + Resp);
                    AddLogLine (AppClientTitle + ': Waiting for Windows Service to Stop');
                    Exit;
                end;
            end;

        // wait for Windows service to stop
        // if does not stop, crash program by process ID or WM_QUIT message
            if AppStopState = StopStateStopWait then begin
                if NOT Running then begin
                    AppStopState := StopStateStopped;
                    AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                    AddLogLine (AppClientTitle + ': Windows Service Has Stopped');
                    AppWinPID := 0;
                    AppWinHnd := 0;
                    Exit;
                end
                else begin
                    if StoppedSecs > FStopWaitSecs then begin
                        AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                        AppStopState := StopStateCrashProg;
                        if AppWinPID <> 0 then begin
                            AddLogLine (AppClientTitle + ': Service Failed to Stop, Terminating Program by Process ID');
                            if NOT MagTermPID (AppWinPID, exitcode) then
                                AddLogLine (AppClientTitle + ': Failed to Terminate Program - ' + MagFormatLastError);
                        end
                        else if AppWinHnd <> 0 then begin
                            AddLogLine (AppClientTitle + ': Service Failed to Stop, Terminating Program by QUIT Message');
                            PostMessage (AppWinHnd, WM_QUIT, 0, 0) ;
                        end
                        else begin
                            AddLogLine (AppClientTitle + ': Failed to Stop Program, No Handles');
                            AppStopState := StopStateNone;
                            Exit;
                        end;
                        AppStopState := StopStateCrashWait;
                        Exit;
                    end;
                end;
            end;

        // still waiting for Windows service to stop
            if AppStopState = StopStateCrashWait then begin
                if NOT Running then begin
                    AppStopState := StopStateStopped;
                    AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                    AddLogLine (AppClientTitle + ': Windows Service Has Stopped');
                    AppWinPID := 0;
                    AppWinHnd := 0;
                    Exit;
                end
                else begin
                    if StoppedSecs > FStopWaitSecs then begin   // give up
                        AddLogLine (AppClientTitle + ': Failed to Stop Windows Service');
                        AppStopState := StopStateNone;
                        Exit;
                    end;
                end;
            end;
        end

    // normal application should be cleanly closed, or crashed
    // beware a Windows service can not easily run a program in the desktop window
        else begin

         // stop program cleanly by WM_CLOSE message to main window, or by process ID if no handle
            if AppStopState = StopStateStopProg then begin
                if Running then begin
                    AddLogLine(AppClientTitle + ': Stopping Windows Program - ' + AppExeFile);
                    if AppWinHnd <> 0 then begin
                        AddLogLine (AppClientTitle + ': Stopping Windows Program by CLOSE Message');
                        PostMessage (AppWinHnd, WM_CLOSE, 0, 0) ;
                    end
                    else if AppWinPID <> 0 then begin
                        AddLogLine (AppServiceName + ' : Stopping Program by Windows Process ID');
                        if NOT MagTermPID (AppWinPID, exitcode) then
                            AddLogLine (AppClientTitle + ': Failed to Terminate Windows Program - ' + MagFormatLastError);
                    end
                    else begin
                        AddLogLine (AppClientTitle + ': Failed to Stop Windows Program, No Handles');
                        AppStopState := StopStateNone;
                        Exit;
                    end;
                    AddLogLine (AppClientTitle + ': Waiting for Windows Program to Stop');
                end;
                AppStopState := StopStateStopWait;
                Exit;
            end;

        // wait for procgram to stop
        // if does not stop, crash program by process ID or WM_QUIT message
            if AppStopState = StopStateStopWait then begin
                if NOT Running then begin
                    AppStopState := StopStateStopped;
                    AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                    AddLogLine (AppClientTitle + ': Windows Program Has Stopped');
                    AppWinPID := 0;
                    AppWinHnd := 0;
                    Exit;
                end
                else begin
                    if StoppedSecs > FStopWaitSecs then begin
                        AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                        AppStopState := StopStateCrashProg;
                        if AppWinPID <> 0 then begin
                            AddLogLine (AppClientTitle + ': Windows Program Failed to Stop, Terminating Program by Process ID');
                            if NOT MagTermPID (AppWinPID, exitcode) then
                                AddLogLine (AppClientTitle + ': Failed to Terminate Windows Program - ' + MagFormatLastError);
                        end
                        else if AppWinHnd <> 0 then begin
                            AddLogLine (AppClientTitle + ': Windows Program Failed to Stop, Terminating Program by QUIT Message');
                            PostMessage (AppWinHnd, WM_QUIT, 0, 0) ;
                        end;
                        AppStopState := StopStateCrashWait;
                        Exit;
                    end;
                end;
            end;

        // still waiting for program to stop
            if AppStopState = StopStateCrashWait then begin
                if NOT Running then begin
                    AppStopState := StopStateStopped;
                    AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                    AddLogLine (AppClientTitle + ': Windows Program Has Stopped');
                end
                else begin
                    if StoppedSecs > FStopWaitSecs then begin   // give up
                        AddLogLine (AppClientTitle + ': Failed to Stop Windows Program');
                        Exit;
                    end;
                end;
            end;
        end;
        except
            AddDiagLine ('Exception AppStopOne: ' + IcsGetExceptMess (ExceptObject)) ;
            AppStopState := StopStateNone;
        end;
    end;
end;


// start a Windows service or program, always exit after changing State
procedure TMonitorMain.AppStartOne(AppNr: Integer);
var
    Resp: String;
    StoppedSecs: Integer;
    Running: Boolean;
    ProcInfo: TProcessInformation;
begin
    if FStoppingFlag then
       Exit;
    if AppNr >= FMaxSuppApps then
        Exit;
    Running := IsAppRunning(AppNr);
    with FMonSuppApp[AppNr] do begin
    try
        StoppedSecs := IcsElapsedSecs64 (AppStoppingTick);
        if AppExeFile = '' then begin
            AddLogLine (AppClientTitle + ': Hello Failed, No EXE Filer');
            AppStartState := StartStateNone;
            Exit;
        end;

     // service needs to be stopped, program confirmed as not running, then service started again
     // beware, this program needs admin rights to start and stop Windows services, not under the debugger
        if FIsAdmin and (AppServiceName <> '') and (AppServiceName <> ServiceNone) then begin
            if FServiceCDatabaseHandle = 0 then begin
                AddLogLine (AppClientTitle + ': No Service Control Handle');
                AppStartState := StartStateNone;
                Exit;
            end;

        // maybe wait 15s before starting Windows service, it may not have closed immediately, and may start itself
            if AppStartState = StartStateStartWait then begin
                if Running then begin
                    AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                            ': Windows Service Already Restarted - ' + AppServiceName + IcsCRLF +
                            'Reason: ' + AppRestartInfo + IcsCRLF + IcsCRLF;
                    AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    AppStartState := StartStateProgStarted;
                    Exit;
                end;
                if StoppedSecs > FStartWaitSecs then begin
                    AppStartState := StartStateProgStarting;
                    AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                end;
            end;

        // now start Windows service again
            if AppStartState = StartStateProgStarting then begin
                AddLogLine (AppClientTitle + ': Starting Windows Service - ' +  AppServiceName);
                AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                            ': Starting Windows Service - ' + AppServiceName + IcsCRLF +
                            'Reason: ' + AppRestartInfo + IcsCRLF + IcsCRLF;
             // no wait or error checking, keep checking if service starts, don't want to block this app
                MagCrtlServEx (FServiceCDatabaseHandle, AppServiceName, servfStart, 0, Resp);
                AddLogLine (AppServiceName + ' : ' + Resp);
                AppStartState := StartStateProgStarted;
                AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                Exit;
            end;

        // wait and check Windows service is actually running
        // BEWARE, may never reach here if client socket reconnects and starts new monitoring session, resettings states
            if AppStartState = StartStateProgStarted then begin
                if Running then begin
                    AddLogLine (AppClientTitle + ': Started Windows Service OK - ' +  AppServiceName);
                    AppEmailLines := AppEmailLines  + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                            ': Restarted Windows Service OK - ' + AppServiceName + IcsCRLF;
                    AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    AppStartState := StartStateProgStarted;
                    Exit;
                end
                else begin
                    if StoppedSecs > FStopWaitSecs then begin   // give up
                        AddLogLine (AppClientTitle + ': Unable to Confirm Service Started');
                        AppEmailLines := AppEmailLines  + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                ': Failed to Restart Windows Service - ' + AppServiceName + IcsCRLF +
                                'Reason: ' + AppRestartInfo + IcsCRLF + IcsCRLF;
                        AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                        AppStartState := StartStateNone;
                        Exit;
                    end;
                end;
            end;

        // sanity test, already running, nothing to do
            if Running then begin
                AddLogLine(AppClientTitle + ': Windows Service Aleady Running - ' +  AppServiceName);
                AppStartState := StartStateNone;
                Exit;
            end;
        end

    // normal application should be cleanly closed, or crashed, then run again
        else begin

         // stop program cleanly by WM_CLOSE message to main window, or by process ID if no handle
            if (MagGetExePID(AppExeFile) <> 0) then begin   // check process lists to see if our EXE is runnng, don't have handle or PID
                AddLogLine(AppClientTitle + ': Windows Program Aleady Running - ' +  AppExeFile);
                AppStartState := StartStateNone;
                Exit;
            end;

        // wait 15s before starting program, it may not have closed immediately, and may restart itself
            if AppStartState = StartStateStartWait then begin
                if (MagGetExePID(AppExeFile) <> 0) then begin   // check process lists to see if our EXE is runnng, don't have handle or PID
                    AppEmailLines := AppEmailLines  + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                             ': Windows Program Already Restarted - ' + AppExeFile + IcsCRLF +
                             'Reason: ' + AppRestartInfo + IcsCRLF + IcsCRLF;
                    AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    AppStartState := StartStateProgStarted;
                    Exit;
                end;
                if StoppedSecs > FStartWaitSecs then begin
                    AppStartState := StartStateProgStarting;
                    AppStoppingTick := IcsGetTickCount64;  // reset stop timer
                end;
            end;

        // now start program again
            if AppStartState = StartStateProgStarting then begin
                AddLogLine (AppClientTitle + ': Starting Windows Program - ' +  AppExeFile);
                ProcInfo := MagStartExe (AppExeFile, '', SW_SHOWNORMAL);
                if ProcInfo.hProcess <> 0 then begin
                    AppWinPID := ProcInfo.dwProcessId;
                    AddLogLine (AppClientTitle + ': Started Windows Program OK - ' +  AppExeFile + ', PID=' + IntToStr(AppWinPID));
                    AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                            ': Restarted Windows Program OK - ' + AppExeFile + IcsCRLF +
                            'Reason: ' + AppRestartInfo + IcsCRLF + IcsCRLF;
                    AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    AppStartState := StartStateProgStarted;
                    Exit;
                end
                else begin
                    AddLogLine (AppClientTitle + ': Failed to Run Windows Program - ' +  AppExeFile);
                    AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                            ': Failed to Started Windows Program - ' + AppExeFile + IcsCRLF +
                            'Reason: ' + AppRestartInfo + IcsCRLF + IcsCRLF;
                    AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    AppStartState := StartStateNone;
                    Exit;
                end;
            end
        end;
        except
            AddDiagLine ('Exception AppStartOne: ' + IcsGetExceptMess (ExceptObject)) ;
            AppStartState := StartStateNone;
        end;
    end;
end;


procedure TMonitorMain.TimerMainTimer(Sender: TObject);
var
    Packet: String;
    AppNr, DeadSecs: Integer;
begin
    TimerMain.Enabled := False;
    try
        try
            if IcsTestTrgTick64(FTriggerBcast) then begin
                FTriggerBcast := IcsGetTrgSecs64 (FBCastSec) ;
                SendBroadcast;
            end;

           if IcsTestTrgTick64(FTriggerMinute) then begin
                FTriggerMinute := IcsGetTrgSecs64 (60) ;
              // update date in registry
                if FIsAdmin then
                    IcsRegPutHlm(IcsAppMonRegKey, IcsAppMonRegDT, RFC3339_DateToStr(Now));  // ISO date
            end;

        // check all applications once a second
            if FTotSuppApps <> 0 then begin
                for AppNr := 0 to FMaxSuppApps - 1 do begin
                     with FMonSuppApp[AppNr] do begin

                     // monitoring has stopped, send pending emails, application might connect again if it restarts
                        if (AppSuppState in [AppStateNone, AppStateStopped]) then begin
                            if AppEmailLines <> '' then begin
                                AddLogLine(AppClientTitle + ': Queuing Email');
                                SendAdminEmail (AdminEmailTo, AppEmailTitle, AppEmailLines);
                                AppEmailLines := '';
                            end;

                           // deallocate record after x minutes??
                            //
                            Continue;
                        end;

                     // stopping monitoring, close server connection if still open
                     // maybe restart application
                        if AppSuppState = AppStateStopping then begin
                            if (AppClientID > 0) and (MonitorServer.States[AppClientId] = logstateOK) then begin
                               AddLogLine(AppClientTitle + ': Disconnecting Client');
                               MonitorServer.CloseSrvCli(AppClientId);
                               AppClientID := 0; // long longer available
                            end;
                            AddLogLine(AppClientTitle + ': Stopped Monitoring, No Client Restart');
                            if AppEmailLines <> '' then begin
                                AddLogLine(AppClientTitle + ': Queuing Email');
                                SendAdminEmail (AdminEmailTo, AppEmailTitle, AppEmailLines);
                            end;
                            AppSuppState := AppStateStopped;
                            AppLastStartTick := 0;
                            AppStopped := Now;
                            Continue;
                        end;

                     // stop an application, note this state remains until application finally stops, couple of minutes maybe
                        if AppSuppState = AppStateProgStop then begin
                           if AppRestartMode = RestModeNot then begin    // sanity check, must not restart
                                AppSuppState := AppStateStopping;
                                Continue;
                           end;

                        // first time here, StopState controls what happens next
                            if AppStopState = StopStateNone then begin
                                AddLogLine(AppClientTitle + ': Stopping Client');
                                AppStopState := StopStateStopProg;
                                AppStoppingTick := IcsGetTickCount64;
                            end;

                        // handle stop events, called once a second
                            AppStopOne(AppNr);

                          // if application has been stopped, see if restarting it
                            if (AppStopState in [StopStateNone, StopStateStopped]) then begin
                               if (AppRestartMode in [RestModeAuto, RestModeRequest, RestModeItelf]) then begin
                                   AddLogLine(AppClientTitle + ': Restarting Client After ' + IntToStr(FStartWaitSecs) + ' Seconds Wait');
                                   AppSuppState := AppStateProgStart;
                                   AppStartState := StartStateStartWait;  //  wait 15 secs for prog to stop
                               end
                               else
                                    AppSuppState := AppStateStopped;
                                AppStopped := Now;
                            end;
                            Continue;
                        end;

                     // start application, note this state remains until it starts
                        if AppSuppState = AppStateProgStart then begin

                        // first time here, StartState controls what happens next
                            if AppStartState = StartStateNone then begin
                                AddLogLine(AppClientTitle + ': Starting Client');
                                AppStartState := StartStateProgStarting;
                                AppStoppingTick := IcsGetTickCount64;
                            end;

                        // handle start events, called once a second, but stop monitoring once it's running
                            AppStartOne(AppNr);
                            if (AppStartState in [StartStateNone, StartStateProgStarted]) then begin
                                AppSuppState := AppStateStopped;
                                AppStopped := Now;
                            end;
                            Continue;
                        end;

                     // lost monitoring connection with client, can not send any more commands
                     // state should have been changed in event, this is fail safe }
                        if (AppSuppState = AppStateOK) and (AppClientID = 0) and (AppStopState = StopStateNone) then begin
                            AppRestartInfo := 'Monitor Socket Closed';
                            AddLogLine(AppClientTitle + ': Monitor Socket Closed, Restarting');
                            AppSuppState := AppStateProgStop;
                            Continue;
                        end;

                     // acknowledge packet received with PONG back to client
                        if (AppSuppState = AppStateOK) and (AppLastPingTick > AppLastPongTick) then begin
                            AppLastPongTick := IcsGetTickCount64;
                            Packet := IcsAppMonName + ProtDelim +
                                      ProtCmdPong + ProtDelim +
                                      AppClientTitle + ProtDelim;
                            if DiagLogPackets.Checked then
                                AddLogLine('Sending: ' + Packet);
                            MonitorServer.SendLogLine(Packet, AppClientID);
                            Continue;
                        end;

                    // exceeded our maximum wait time with no PINGs, try and wake up client
                        if AppSuppState = AppStateOK then begin
                            DeadSecs := IcsElapsedSecs64 (AppLastPingTick);
                            if DeadSecs > AppDeadSecs then begin
                                AppSuppState := AppStateQuiet;
                                AppLastPongTick := IcsGetTickCount64;
                                Packet := IcsAppMonName + ProtDelim +
                                          ProtCmdWake + ProtDelim +
                                          AppClientTitle + ProtDelim;
                                AddLogLine('Client gone QUIET, Sending: ' + Packet);
                                MonitorServer.SendLogLine(Packet, AppClientID);
                                Continue;
                            end;
                        end;

                    // wait 30 seconds for response to WAKE, otherwsie stop and restart application
                        if AppSuppState = AppStateQuiet then begin
                            DeadSecs := IcsElapsedSecs64 (AppLastPongTick);
                            if DeadSecs > AppDeadSecs then begin
                                AddLogLine(AppClientTitle + ': Client Still Quiet After ' + IntToStr(DeadSecs) + ', Restarting');
                                AppSuppState := AppStateProgStop;
                                AppRestartInfo := 'Client Stopped Sending Pings';
                                Continue;
                            end;
                        end;

                     // if application has stopped on it's own, try and restart it
                     // should have lost monitor socket already
                        if AppSuppState = AppStateOK then begin
                            if NOT IsAppRunning(AppNr) then begin
                                AddLogLine(AppClientTitle + ': Program No Longer Running, Restarting');
                                AppRestartInfo := 'Program Not Running';
                                AppSuppState := AppStateProgStop;
                                Continue;
                            end;
                        end;

                     // see if pending email, wait 30 seconds for more lines to be added
                        if AppEmailTick <> Trigger64Disabled then begin
                            if IcsElapsedSecs64 (AppEmailTick) > 30 then begin
                                AppEmailTick := Trigger64Disabled;
                                if AppEmailLines <> '' then begin
                                    AddLogLine(AppClientTitle + ': Queuing Email');
                                    SendAdminEmail (AdminEmailTo, AppEmailTitle, AppEmailLines);
                                end;
                                AppEmailLines := '';
                            end;
                        end;

                     end;
                end;
            end;
        except
            AddDiagLine ('Exception Main Timer: ' + IcsGetExceptMess (ExceptObject)) ;
        end;
    finally
        TimerMain.Enabled := True;
    end;
end;


procedure TMonitorMain.TriggerMessageEvent (var Msg: TMsg; var Handled: Boolean) ;
begin

// a client is asking us to broadcast our HELLO message immediately
    if (Msg.message = FWinMess_Query) and (Msg.wParam = IcsAppMonMessParam) then begin
    //    AddDiagLine ('WinMess_QueryInfo: ' + IntToStr(Msg.wParam)) ;
        if FTriggerBcast <> Trigger64Immediate then begin  // try and stop multiple loggings
            AddDiagLine ('WinMess_QueryInfo: ' + IntToStr(Msg.wParam)) ;
            FTriggerBcast := Trigger64Immediate;
        end;
        Handled := true ;
    end;

// see if system is closing down
    if (Msg.message = WM_QUERYENDSESSION) then begin
        AddDiagLine ('!! Windows Wants to Close Down, QuerySessionEnd') ;
    end;
    if (Msg.message = WM_ENDSESSION) then begin
        AddDiagLine ('!! Windows Is Closing Down, Stop Monitoring SessionEnd') ;
        MonStop;
    end;
end;


procedure TMonitorMain.MonitorBroadcastLogProgEvent(Sender: TObject; Socnr: Integer; LogOption: TLogOption; const Msg: string);
begin
    AddLogLine ('Broadcast Event: ' + Msg) ;   // ignore it
end;


procedure TMonitorMain.MonitorBroadcastLogRecvEvent(Sender: TObject; Socnr: Integer; const Line: string);
begin
    AddLogLine ('Broadcast response from: ' + MonitorBroadcast.Socket[SocNr].PeerAddr + ' - ' + Line);
end;

procedure TMonitorMain.HardwareEventsPowerEvent(Sender: TObject; PowerEvent: Integer; const Desc: string);
begin
    AddLogLine ('System Power Event: ' + Desc + ', Event=' + IntToStr(PowerEvent));
//    if PowerEvent += PBT_ then

end;


procedure TMonitorMain.MonitorServerLogChangeEvent(Sender: TObject; Socnr: Integer; LogState: TStrmLogState);
var
    AppNr: Integer;
    Title: String;
begin
    AppNr := -1;
    Title := 'Unknown';
    if (SocNr > 0) then begin
        AppNr := FindSuppApp(SocNr);
        if (AppNr >= 0) and (LogState >= logstateOK) then
             Title := FMonSuppApp[AppNr].AppClientTitle;
        AddLogLine (Title + ': State ' + StrmLogStateNames[LogState] + ' CliId ' + IntToStr(SocNr)) ;
    end;
    if  (AppNr >= 0) and (LogState = logstateStopping) then begin
        with FMonSuppApp[AppNr] do begin
            if (AppRestartMode = RestModeNot) or FStoppingFlag then begin
                AddLogLine(AppClientTitle + ': Stopped Application Monitoring');
                AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                                ': Stopped Application Monitoring' + IcsCRLF + IcsCRLF;
                AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                AppSuppState := AppStateStopped;
                AppLastStartTick := 0;
                AppStopped := Now;
            end
            else begin
                if AppSuppState = AppStateOK then begin
                    AppSuppState := AppStateProgStop;
                    AppRestartInfo := 'Monitor Socket Closed Unexpectedly';
                    AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                            ': Monitor Socket Closed Unexpectedly, Restarting' + IcsCRLF + IcsCRLF;
                    AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    AddLogLine(AppClientTitle + ': Monitor Socket Closed Unexpectedly, Restarting');
                    AppLastStartTick := 0;
                    AppClientID := 0; // long longer available
                end;
            end;
        end;
    end;
end;


procedure TMonitorMain.MonitorServerLogProgEvent(Sender: TObject; Socnr: Integer; LogOption: TLogOption; const Msg: string);
var
    AppNr: Integer;
    Title: String;
begin
    Title := 'New Client';
    if (SocNr > 0) then begin
        AppNr := FindSuppApp(SocNr);
        if (AppNr >= 0) and (FMonSuppApp[AppNr].AppSuppState >= AppStateOK) then
             Title := FMonSuppApp[AppNr].AppClientTitle;
        AddLogLine(Title + ': ' + Msg + ' CliId ' + IntToStr(SocNr))
    end
    else
        AddLogLine (Msg);
end;

procedure TMonitorMain.MonitorServerLogRecvEvent(Sender: TObject; Socnr: Integer; const Line: string);
var
    AppNr: Integer;
    CliMode: TAMMode;
    EmailBody, Command, Packet, CliTitle: String;

    function FindAMMode(const Lit: String): TAMMode;
    var
        Mode: TAMMode;
    begin
        Result := AMModeMonitor;
        for Mode := Low(TAMMode) to High(TAMMode) do begin
            if Lit = AMModeCmds[Mode] then begin
                Result := Mode;
                Exit;
            end;
        end;
    end;

    function FindAMRestart(const Lit: String): TAMRestartMode;
    var
        Mode: TAMRestartMode;
    begin
        Result := RestModeNone;
        for Mode := Low(TAMRestartMode) to High(TAMRestartMode) do begin
            if Lit = AMRestModeCmds[Mode] then begin
                Result := Mode;
                Exit;
            end;
        end;
    end;

    function FindOldApp: Integer;
    var
        I: Integer;
    begin
        Result := -1;
        if FTotSuppApps = 0 then
            Exit;
        for I := 0 to FMaxSuppApps - 1 do begin
            if (FMonSuppApp[I].AppClientTitle = CliTitle) and (FMonSuppApp[I].AppMonMode = CliMode) then begin
                Result := I;
                exit;
            end;
        end;
    end;

begin
    if DiagLogPackets.Checked then
        AddLogLine ('Server message from ClientId: ' + IntToStr(SocNr) + ' - ' + Line) ;
    FAppFields.Clear;
    EmailBody := '';
    FAppFields.DelimitedText := Line;
    try
        if (FAppFields.Count > CliCFieldLast) then begin

         // sanity checking
            if FAppFields[CliCFieldAppMon] = IcsAppMonName then begin
                CliTitle := FAppFields[CliCFieldAppTitle];

             // find AppNr for current session from SocNr
                AppNr := FindSuppApp(SocNr);
                if Appnr < 0 then begin

             // find AppNr for old session from AppTitle and Mode
                    CliMode := AMModeMonitor;
                    if (FAppFields.Count > CliHFieldDeadSecs) then    // assume HELLO
                        CliMode := FindAMMode(FAppFields[CliHFieldAppMode]);
                    Appnr := FindOldApp;
                    if Appnr < 0 then begin

             // app not connected before, new record
                        AppNr := AddSuppApp;
                        AddLogLine('Allocate New App Record ' + IntToStr(AppNr + 1));
                   end
                   else begin
                        AddLogLine('Using Old App Record ' + IntToStr(AppNr + 1));
                        with FMonSuppApp[AppNr] do begin
                            if (AppStartState > StartStateNone) then begin  // were we restarting application?
                                AddLogLine (AppClientTitle + ': Application Restarted OK');
                                AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                                            ': Application Restarted OK' + IcsCRLF + IcsCRLF;
                                AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                            end;
                        end;
                   end;
                   FMonSuppApp[AppNr].AppClientId := SocNr;
                end;
                if Appnr < 0 then begin
                    AddLogLine('Failed to Allocate New Record');
                    exit;
                end;

            // handle received commands from clients
               with FMonSuppApp[AppNr] do begin
                    Command := FAppFields[CliCFieldCommand];
                    EmailBody := IcsUnEscProtField(FAppFields[CliCFieldEmail]);
                    if EmailBody <> '' then begin
                        AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + ': ' + EmailBody + IcsCRLF + IcsCRLF;
                        AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                    end;

                // HELLO command, remote client is introducing itself
                    if (Command = ProtCmdHello) and (FAppFields.Count > CliHFieldLast)  then begin
                        try
                            AppClientId := SocNr;
                            AppClientTitle := CliTitle;
                            AppMonMode := CliMode;
                            AppRestartMode := RestModeNone;
                            AppStopState := StopStateNone;
                            AppStartState := StartStateNone;
                         // NonStop mode means we automatically restart unless told otherwise
                            if AppMonMode = AMModeNonStop then
                                AppRestartMode := RestModeAuto;
                            AppExeFile := Trim(FAppFields[CliHFieldExeFile]);
                            AppServiceName := Trim(FAppFields[CliHFieldServiceName]);
                            if AppServiceName = '' then
                               AppServiceName := ServiceNone;
                            AppWinHnd := atoi64(FAppFields[CliHFieldWinHnd]);
                            AppWinPID := atoi64(FAppFields[CliHFieldWinPID]);
                            AppCompName := Trim(FAppFields[CliHFieldCompName]);
                            AppLocalIP := Trim(FAppFields[CliHFieldLocalIP]);
                            AppDeadSecs := atoi(FAppFields[CliHFieldDeadSecs]);
                            AppGeneral := IcsUnEscProtField(FAppFields[CliHFieldAppGeneral]);
                            AppStatus := '';
                            if AppDeadSecs < 30 then
                                AppDeadSecs := FDeadSecs;
                            AppStarted := Now;
                            AppLastStartTick := IcsGetTickCount64;
                            AppSuppState := AppStateOK;
                            AppLastPingTick := IcsGetTickCount64;
                            if AppLocalIP = '' then
                            try
                                AppLocalIP := MonitorServer.SrvClientSoc[SocNr].CPeerAddr;
                            except
                                AddLogLine('Failed to get CPeerAddr');
                            end;
                            AppEmailTitle := AppClientTitle + ' on ' + AppCompName;
                            if AppServiceName <> ServiceNone then
                                AppEmailTitle := AppEmailTitle + ', Service: ' + AppServiceName;

                        // see if unable to restart application
                            if AppMonMode = AMModeNonStop then begin
                                if (NOT FIsAdmin) and (AppServiceName <> ServiceNone) then begin
                                    AddLogLine(AppClientTitle + ': Need Admin Rights to Control Windows Services: ' + AppServiceName);
                                    Packet := IcsAppMonName + ProtDelim +
                                              ProtCmdDecline + ProtDelim +
                                              AppClientTitle + ProtDelim +
                                              ProtDelim;
                                    AddLogLine(AppClientTitle + ': Declined Monitoring, Sending: ' + Packet);
                                    MonitorServer.SendLogLine(Packet, AppClientID);
                                    AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                                                      ': Declined Monitoring' + IcsCRLF + IcsCRLF;
                                    AppRestartMode := RestModeNot;
                                    AppSuppState := AppStateStopping;
                                    Exit;
                                end;
                                if GStartedByScm and (AppServiceName = ServiceNone) then begin
                                    AddLogLine(AppClientTitle + ': Windows Services Can Not Run Desktop Programs');
                                    Packet := IcsAppMonName + ProtDelim +
                                              ProtCmdDecline + ProtDelim +
                                              AppClientTitle + ProtDelim +
                                              ProtDelim;
                                    AddLogLine(AppClientTitle + ': Declined Monitoring, Sending: ' + Packet);
                                    MonitorServer.SendLogLine(Packet, AppClientID);
                                    AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                                                      ': Declined Monitoring' + IcsCRLF + IcsCRLF;
                                    AppRestartMode := RestModeNot;
                                    AppSuppState := AppStateStopping;
                                    Exit;
                                end;
                            end;
                            AddLogLine(AppClientTitle + ': Application Monitoring Started: ' + Copy(EmailBody, 1, 80));

                            if IsAppRunning(AppNr) then
                                AddLogLine(AppClientTitle + ': ' + AppExeFile + ' is Running OK')
                            else begin
                                if AppServiceName = ServiceNone then
                                   AddLogLine(AppClientTitle + ': ' + AppExeFile + ' Not Found Running!!! PID=' + IntToStr(AppWinPID));
                            end;

                        // email, may have message from app as well
                            AppEmailLines := AppEmailLines + MagDateTimeToAStr(Now) + IcsSpace + AppClientTitle +
                                       ': Application Monitoring Started' + IcsCRLF + 'File : ' + AppExeFile + IcsCRLF + IcsCRLF;
                            AppEmailTick := IcsGetTickCount64;  // sends email in 30 seconds
                        except
                            AddDiagLine (CliTitle + ': Exception HELLO Packet: ' + IcsGetExceptMess (ExceptObject)) ;
                            AppRestartMode := RestModeNot;
                            AppSuppState := AppStateStopping;
                        end;
                    end

                 // PING command, expected regularly from remote server
                    else if Command = ProtCmdPing then begin
                        try
                            if (AppLastStartTick = 0) or (AppExeFile = '') then begin
                                AppClientId := SocNr;
                                AddLogLine(AppClientTitle + ': Non-Registered Application, Closing');
                                AppRestartMode := RestModeNot;
                                AppSuppState := AppStateStopping;
                                Exit;
                            end;
                            if EmailBody <> '' then
                                AddLogLine(AppClientTitle + ': PING and Email Received from Client: ' + Copy(EmailBody, 1, 80))
                            else
                                AddLogLine(AppClientTitle + ': PING Packet Received from Client');
                            if (FAppFields.Count > CliPFieldStatus) then
                                AppStatus := IcsUnEscProtField(FAppFields[CliPFieldStatus]);
                            AppLastPingTick := IcsGetTickCount64;
                            AppSuppState := AppStateOK;
                        except
                            AddDiagLine (CliTitle + ': Exception PING Packet: ' + IcsGetExceptMess (ExceptObject)) ;
                            AppRestartMode := RestModeNot;
                            AppSuppState := AppStateStopping;
                        end;
                    end

                 // STOP command, remote server wants to stop being monitored, it is probably stopping
                 // generally don't restart unless a RESTART command has been received already
                    else if (Command = ProtCmdStop) and (FAppFields.Count > CliSFieldRestart) then begin
                        try
                            if EmailBody <> '' then
                                AddLogLine(AppClientTitle + ': STOP and Email Received from Client: ' + Copy(EmailBody, 1, 80))
                            else
                                AddLogLine(AppClientTitle + ': STOP Packet Received');
                            if AppSuppState >= AppStateStopping then   // duplicate or restarting
                                Exit;
                            AppLastPingTick := IcsGetTickCount64;
                            if MonitorServer.States[SocNr] = logstateOK then begin
                                Packet := IcsAppMonName + ProtDelim +
                                          ProtCmdPong + ProtDelim +
                                          AppClientTitle + ProtDelim +
                                          ProtDelim;
                                AddLogLine(AppClientTitle + ': Acknowledged STOP command: ' + Packet);
                            end;
                            MonitorServer.SendLogLine(Packet, AppClientID);

                        // NONESTOP mode, see what are doing
                            if AppMonMode = AMModeNonStop then begin
                                AppRestartMode := FindAMRestart(FAppFields[CliSFieldRestart]);
                                AddLogLine(AppClientTitle + ': Control Requested: ' + AMRestModeNames[AppRestartMode]);
                            end
                            else
                                AppRestartMode := RestModeNot;
                            AppRestartInfo := AMRestModeNames[AppRestartMode];
                            case AppRestartMode of
                                RestModeNone: AppSuppState := AppStateStopping;
                                RestModeAuto: AppSuppState := AppStateProgStop;
                                RestModeRequest: AppSuppState := AppStateProgStop;
                                RestModeItelf: begin
                                                  AppSuppState := AppStateProgStop;  // don't stop, but check it does stop
                                                  AppStopState := StopStateStopWait;
                                               end;
                                RestModeNot: AppSuppState := AppStateStopping;
                            end;
                            AppStoppingTick := IcsGetTickCount64;
                        except
                            AddDiagLine (CliTitle + ': Exception STOP Packet: ' + IcsGetExceptMess (ExceptObject)) ;
                            AppRestartMode := RestModeNot;
                            AppSuppState := AppStateStopping;
                        end;
                    end

                 // USERINFO command, might have EmailBody and will send email, maybe app wants it?
                    else if Command = ProtCmdUserInfo then begin
                        AddLogLine(AppClientTitle + ': USERINFO Packet Received - ' + Line);
                        AppLastPingTick := IcsGetTickCount64;
                        AppSuppState := AppStateOK;
                     // onUserInfoEvent(Line)
                     // pending
                    end

                 // INSTALL command, replace running EXE with new one, stopping and restarting
                    else if Command = ProtCmdUserInfo then begin
                        AddLogLine(AppClientTitle + ': INSTALL Packet Received - ' + Line);
                        AppLastPingTick := IcsGetTickCount64;
                        AppSuppState := AppStateOK;
                     // onxxerInfoEvent(Line)
                     // pending
                    end

                    else begin
                        AddLogLine('Unexpected Server Packet, Bad Command - ' + Line);
                    end;
                end;
            end
            else
                AddLogLine('Unexpected Server Packet, Bad Command - ' + Line);
        end
        else
             AddLogLine('Unexpected Server Packet, Too Few Fields - ' + Line);
    except
        AddDiagLine ('Exception Processing Packet: ' + IcsGetExceptMess (ExceptObject)) ;
    end;
end;


end.
